#!/usr/bin/env python
# encoding: utf-8

from __future__ import absolute_import, unicode_literals
from .json_exporter import JSONExporter
from .util import get_tags_count
from ..util import u
from xml.dom import minidom


class XMLExporter(JSONExporter):
    """This Exporter can convert entries and journals into XML."""
    names = ["xml"]
    extension = "xml"

    @classmethod
    def export_entry(cls, entry, doc=None):
        """Returns an XML representation of a single entry."""
        doc_el = doc or minidom.Document()
        entry_el = doc_el.createElement('entry')
        for key, value in cls.entry_to_dict(entry).items():
            elem = doc_el.createElement(key)
            elem.appendChild(doc_el.createTextNode(u(value)))
            entry_el.appendChild(elem)
        if not doc:
            doc_el.appendChild(entry_el)
            return doc_el.toprettyxml()
        else:
            return entry_el

    @classmethod
    def entry_to_xml(cls, entry, doc):
        entry_el = doc.createElement('entry')
        entry_el.setAttribute('date', entry.date.isoformat())
        if hasattr(entry, "uuid"):
            entry_el.setAttribute('uuid', u(entry.uuid))
        entry_el.setAttribute('starred', u(entry.starred))
        entry_el.appendChild(doc.createTextNode(entry.fulltext))
        return entry_el

    @classmethod
    def export_journal(cls, journal):
        """Returns an XML representation of an entire journal."""
        tags = get_tags_count(journal)
        doc = minidom.Document()
        xml = doc.createElement('journal')
        tags_el = doc.createElement('tags')
        entries_el = doc.createElement('entries')
        for count, tag in tags:
            tag_el = doc.createElement('tag')
            tag_el.setAttribute('name', tag)
            count_node = doc.createTextNode(u(count))
            tag_el.appendChild(count_node)
            tags_el.appendChild(tag_el)
        for entry in journal.entries:
            entries_el.appendChild(cls.entry_to_xml(entry, doc))
        xml.appendChild(entries_el)
        xml.appendChild(tags_el)
        doc.appendChild(xml)
        return doc.toprettyxml()



#!/usr/bin/env python
# encoding: utf-8

from __future__ import absolute_import, unicode_literals
from .util import ERROR_COLOR, RESET_COLOR
from .util import slugify, u
from .template import Template
import os
import codecs


class Exporter(object):
    """This Exporter can convert entries and journals into text files."""
    def __init__(self, format):
        with open("jrnl/templates/" + format + ".template") as f:
            front_matter, body = f.read().strip("-\n").split("---", 2)
            self.template = Template(body)

    def export_entry(self, entry):
        """Returns a unicode representation of a single entry."""
        return entry.__unicode__()

    def _get_vars(self, journal):
        return {
            'journal': journal,
            'entries': journal.entries,
            'tags': journal.tags
        }

    def export_journal(self, journal):
        """Returns a unicode representation of an entire journal."""
        return self.template.render_block("journal", **self._get_vars(journal))

    def write_file(self, journal, path):
        """Exports a journal into a single file."""
        try:
            with codecs.open(path, "w", "utf-8") as f:
                f.write(self.export_journal(journal))
                return "[Journal exported to {0}]".format(path)
        except IOError as e:
            return "[{2}ERROR{3}: {0} {1}]".format(e.filename, e.strerror, ERROR_COLOR, RESET_COLOR)

    def make_filename(self, entry):
        return entry.date.strftime("%Y-%m-%d_{0}.{1}".format(slugify(u(entry.title)), self.extension))

    def write_files(self, journal, path):
        """Exports a journal into individual files for each entry."""
        for entry in journal.entries:
            try:
                full_path = os.path.join(path, self.make_filename(entry))
                with codecs.open(full_path, "w", "utf-8") as f:
                    f.write(self.export_entry(entry))
            except IOError as e:
                return "[{2}ERROR{3}: {0} {1}]".format(e.filename, e.strerror, ERROR_COLOR, RESET_COLOR)
        return "[Journal exported to {0}]".format(path)

    def export(self, journal, format="text", output=None):
        """Exports to individual files if output is an existing path, or into
        a single file if output is a file name, or returns the exporter's
        representation as unicode if output is None."""
        if output and os.path.isdir(output):  # multiple files
            return self.write_files(journal, output)
        elif output:                          # single file
            return self.write_file(journal, output)
        else:
            return self.export_journal(journal)



#!/usr/bin/env python
# encoding: utf-8


"""
jrnl is a simple journal application for your command line.
"""
from __future__ import absolute_import

__title__ = 'jrnl'
__version__ = 'v2.0.1'
__author__ = 'Manuel Ebert'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2013 - 2015 Manuel Ebert'



from . import Journal, util
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
import hashlib
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend
import sys
import os
import base64
import getpass
import logging

log = logging.getLogger()


def make_key(password):
    password = util.bytes(password)
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        # Salt is hard-coded
        salt=b'\xf2\xd5q\x0e\xc1\x8d.\xde\xdc\x8e6t\x89\x04\xce\xf8',
        iterations=100000,
        backend=default_backend()
    )
    key = kdf.derive(password)
    return base64.urlsafe_b64encode(key)


class EncryptedJournal(Journal.Journal):
    def __init__(self, name='default', **kwargs):
        super(EncryptedJournal, self).__init__(name, **kwargs)
        self.config['encrypt'] = True

    def open(self, filename=None):
        """Opens the journal file defined in the config and parses it into a list of Entries.
        Entries have the form (date, title, body)."""
        filename = filename or self.config['journal']

        if not os.path.exists(filename):
            password = util.getpass("Enter password for new journal: ")
            if password:
                if util.yesno("Do you want to store the password in your keychain?", default=True):
                    util.set_keychain(self.name, password)
                else:
                    util.set_keychain(self.name, None)
                self.config['password'] = password
                text = ""
                self._store(filename, text)
                util.prompt("[Journal '{0}' created at {1}]".format(self.name, filename))
            else:
                util.prompt("No password supplied for encrypted journal")
                sys.exit(1)
        else:
            text = self._load(filename)
        self.entries = self._parse(text)
        self.sort()
        log.debug("opened %s with %d entries", self.__class__.__name__, len(self))
        return self


    def _load(self, filename, password=None):
        """Loads an encrypted journal from a file and tries to decrypt it.
        If password is not provided, will look for password in the keychain
        and otherwise ask the user to enter a password up to three times.
        If the password is provided but wrong (or corrupt), this will simply
        return None."""
        with open(filename, 'rb') as f:
            journal_encrypted = f.read()

        def validate_password(password):
            key = make_key(password)
            try:
                plain = Fernet(key).decrypt(journal_encrypted).decode('utf-8')
                self.config['password'] = password
                return plain
            except (InvalidToken, IndexError):
                return None
        if password:
            return validate_password(password)
        return util.get_password(keychain=self.name, validator=validate_password)

    def _store(self, filename, text):
        key = make_key(self.config['password'])
        journal = Fernet(key).encrypt(text.encode('utf-8'))
        with open(filename, 'wb') as f:
            f.write(journal)

    @classmethod
    def _create(cls, filename, password):
        key = make_key(password)
        dummy = Fernet(key).encrypt(b"")
        with open(filename, 'wb') as f:
            f.write(dummy)


class LegacyEncryptedJournal(Journal.LegacyJournal):
    """Legacy class to support opening journals encrypted with the jrnl 1.x
    standard. You'll not be able to save these journals anymore."""
    def __init__(self, name='default', **kwargs):
        super(LegacyEncryptedJournal, self).__init__(name, **kwargs)
        self.config['encrypt'] = True

    def _load(self, filename, password=None):
        with open(filename, 'rb') as f:
            journal_encrypted = f.read()
        iv, cipher = journal_encrypted[:16], journal_encrypted[16:]

        def validate_password(password):
            decryption_key = hashlib.sha256(password.encode('utf-8')).digest()
            decryptor = Cipher(algorithms.AES(decryption_key), modes.CBC(iv), default_backend()).decryptor()
            try:
                plain_padded = decryptor.update(cipher) + decryptor.finalize()
                self.config['password'] = password
                if plain_padded[-1] in (" ", 32):
                    # Ancient versions of jrnl. Do not judge me.
                    return plain_padded.decode('utf-8').rstrip(" ")
                else:
                    unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
                    plain = unpadder.update(plain_padded) + unpadder.finalize()
                    return plain.decode('utf-8')
            except ValueError:
                return None
        if password:
            return validate_password(password)
        return util.get_password(keychain=self.name, validator=validate_password)



from datetime import datetime
from dateutil.parser import parse as dateparse
try: import parsedatetime.parsedatetime_consts as pdt
except ImportError: import parsedatetime as pdt

FAKE_YEAR = 9999
DEFAULT_FUTURE = datetime(FAKE_YEAR, 12, 31, 23, 59, 59)
DEFAULT_PAST = datetime(FAKE_YEAR, 1, 1, 0, 0)

consts = pdt.Constants(usePyICU=False)
consts.DOWParseStyle = -1  # "Monday" will be either today or the last Monday
CALENDAR = pdt.Calendar(consts)


def parse(date_str, inclusive=False, default_hour=None, default_minute=None):
    """Parses a string containing a fuzzy date and returns a datetime.datetime object"""
    if not date_str:
        return None
    elif isinstance(date_str, datetime):
        return date_str

    # Don't try to parse anything with 6 or less characters. It's probably a markdown footnote
    if len(date_str) <= 6:
        return None

    default_date = DEFAULT_FUTURE if inclusive else DEFAULT_PAST
    date = None
    year_present = False
    while not date:
        try:
            date = dateparse(date_str, default=default_date)
            if date.year == FAKE_YEAR:
                date = datetime(datetime.now().year, date.timetuple()[1:6])
            else:
                year_present = True
            flag = 1 if date.hour == date.minute == 0 else 2
            date = date.timetuple()
        except Exception as e:
            if e.args[0] == 'day is out of range for month':
                y, m, d, H, M, S = default_date.timetuple()[:6]
                default_date = datetime(y, m, d - 1, H, M, S)
            else:
                date, flag = CALENDAR.parse(date_str)

    if not flag:  # Oops, unparsable.
        try:  # Try and parse this as a single year
            year = int(date_str)
            return datetime(year, 1, 1)
        except ValueError:
            return None
        except TypeError:
            return None

    if flag is 1:  # Date found, but no time. Use the default time.
        date = datetime(*date[:3], hour=default_hour or 0, minute=default_minute or 0)
    else:
        date = datetime(*date[:6])

    # Ugly heuristic: if the date is more than 4 weeks in the future, we got the year wrong.
    # Rather then this, we would like to see parsedatetime patched so we can tell it to prefer
    # past dates
    dt = datetime.now() - date
    if dt.days < -28 and not year_present:
        date = date.replace(date.year - 1)
    return date



#!/usr/bin/env python
# encoding: utf-8

from __future__ import absolute_import, unicode_literals

from .text_exporter import TextExporter
from .jrnl_importer import JRNLImporter
from .json_exporter import JSONExporter
from .markdown_exporter import MarkdownExporter
from .tag_exporter import TagExporter
from .xml_exporter import XMLExporter
from .yaml_exporter import YAMLExporter
from .template_exporter import __all__ as template_exporters

__exporters =[JSONExporter, MarkdownExporter, TagExporter, TextExporter, XMLExporter, YAMLExporter] + template_exporters
__importers =[JRNLImporter]

__exporter_types = dict([(name, plugin) for plugin in __exporters for name in plugin.names])
__importer_types = dict([(name, plugin) for plugin in __importers for name in plugin.names])

EXPORT_FORMATS = sorted(__exporter_types.keys())
IMPORT_FORMATS = sorted(__importer_types.keys())

def get_exporter(format):
    for exporter in __exporters:
        if hasattr(exporter, "names") and format in exporter.names:
            return exporter
    return None


def get_importer(format):
    for importer in __importers:
        if hasattr(importer, "names") and format in importer.names:
            return importer
    return None
